import 'package:no_bloc/src/bloc_monitor.dart';
import 'package:meta/meta.dart';
import 'package:rxdart/rxdart.dart';

/// The central class behind this library
///
/// It is responsible for processing all the events and broadcasting corresponding
/// changes made to the state.
/// [R] is the type which subclasses [Bloc]
/// [S] is the type of [value] which this bloc broadcasts. [S] must implement equality
abstract class Bloc<R, S> {
  static bool checkIfValueType = true;
  final BlocMonitor _monitor;

  StateError? _error;
  bool _isBusy = true;
  BehaviorSubject<S?>? _state;
  S? _value;

  /// [initialState] is the state which is set at initialization. if its null, the initial state is set to busy.
  Bloc({S? initialState, BlocMonitor monitor = const BlocEventsPrinter()})
      : _value = initialState,
        _monitor = monitor {
    _monitor.onInit('$R', _value);
    if (initialState != null) assert(isValueType(initialState));
    if (_value != null) {
      _isBusy = false;
    }
  }

  /// Broadcast Stream to which all builders listen
  ///
  /// This stream is created only when there is a listener, and is automatically cleaned
  /// when there are no listeners are left. State change broadcasts are discarded when there
  /// are no listeners.
  Stream<S?> get state {
    _monitor.onStreamListener('$R');
    if (_state == null) {
      _state =
          _value == null ? BehaviorSubject<S>() : BehaviorSubject.seeded(value);

      _state!.onCancel = () {
        if (_state != null && !_state!.hasListener) {
          _monitor.onStreamDispose('$R');
          _state!.close();
          _state = null;
        }
      };
    }

    return _state!.stream;
  }

  S get value {
    if (_value != null) {
      return _value!;
    }

    throw StateError('value cannot be used before it is '
        'set (did you forgot to set initialState)');
  }

  bool get isBusy => _isBusy;

  bool get hasError => _error != null;

  StateError get error => _error!;

  /// Called to calculate the new state
  ///
  /// This function takes [currentState] and the planned [update] state
  /// and returns the state which would be broadcast and will become the current state
  ///
  /// Originally it just discards the current state and returns the [update] as the next state.
  /// This behavior can of course be overridden
  @protected
  S nextState(S? currentState, S update) {
    return update;
  }

  /// Called by blocs (subclasses) when the state is updated
  ///
  /// [isBusy] and [error] are cleared whenever the state is updated
  /// Optional argument [event] is the name of event which is calling [setState]
  @protected
  void setState(S state, {String? event}) {
    assert(state != null);
    assert(isValueType(state));
    final _state = this._state;
    _monitor.onEvent('$R', _value, state, event: event);
    _isBusy = false;
    _error = null;
    final next = nextState(_value, state);

    _value = next;
    if (_state != null) {
      _state.add(_value);
      _monitor.onBroadcast('$R', _value, event: event);
    }
  }

  /// Called by blocs (subclasses) when an error occurs
  ///
  /// [isBusy] is cleared whenever an [error] is set
  /// Optional argument [event] is the name of event which is calling [setError]
  @protected
  void setError(StateError error, {String? event}) {
    if (_error == error) return;
    _monitor.onError('$R', error, event: event);
    _isBusy = false;
    _error = error;
    _state?.add(null);
  }

  /// Called by blocs (subclasses) when the updated state isn't available immediately
  ///
  /// This should be used when a bloc is waiting for an async operation (e.g. a network call)
  /// [error] is cleared whenever [isBusy] is set
  /// Optional argument [event] is the name of event which is calling [setError]
  @protected
  void setBusy({String? event}) {
    if (_isBusy) return;
    _monitor.onBusy('$R', event: event);
    _isBusy = true;
    _error = null;
    _state?.add(null);
  }

  void refresh() {
    _state?.add(_value);
  }

  /// Send a done event and disposes the stream
  ///
  /// It is not necessary to call this function as the stream is disposed automatically.
  /// But is useful for testing, in cases where otherwise the tester keeps waiting until done.
  Future<void> dispose() async {
    if (_state != null) {
      _monitor.onStreamDispose('$R');
      await _state!.close();
    }
  }
}

/// Just a utility function to make sure if the state type implements equality (by default dart classes only support referential equality).
///
/// This library will not work if equality is not implemented. If you are manually overriding == and hashCode for your classes
/// instead of using [Equatable] or [Built], then you have to set [Bloc.checkIfValueType] to false, to avoid getting false errors.
bool isValueType(value) {
  return (!Bloc.checkIfValueType ||
      value is num ||
      value is String ||
      value is DateTime ||
      value is bool);
}
